#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 28 13:46:36 2019

@author: annaquinlan
https://github.com/tidepool-org/LoopKit/blob/
57a9f2ba65ae3765ef7baafe66b883e654e08391/LoopKit/CarbKit/CarbMath.swift
"""
# pylint: disable=R0913, C0200, C0301, R0914, R0915, C0302
import sys
from datetime import timedelta

from pyloopkit.insulin_math import find_ratio_at_time
from pyloopkit.date import (time_interval_since,
                            date_floored_to_time_interval,
                            date_ceiled_to_time_interval)
from pyloopkit import carb_status


def map_(
        carb_entry_starts, carb_entry_quantities, carb_entry_absorptions,
        effect_starts, effect_ends, effect_values,
        carb_ratio_starts, carb_ratios,
        sensitivity_starts, sensitivity_ends, sensitivity_values,
        absorption_time_overrun,
        default_absorption_time,
        delay,
        delta=5
        ):
    """
    Maps a sorted timeline of carb entries to the observed absorbed
    carbohydrates for each, from a timeline of glucose effect velocities.

    This makes some important assumptions:
        - insulin effects, used with glucose to calculate
          counteraction, are "correct"
        - carbs are absorbed completely in the order they were eaten
          without mixing or overlapping effects

    Arguments:
    carb_entry_starts -- list of times of carb entry (datetime objects)
    carb_entry_quantities -- list of grams of carbs eaten
    carb_entry_absorptions -- list of lengths of absorption times (mins)

    effect_starts -- list of start times of carb effect (datetime objects)
    effect_ends -- list of end times of carb effect (datetime objects)
    effect_values -- list of carb effects (mg/dL)

    carb_ratio_starts -- list of start times of carb ratios (time objects)
    carb_ratios -- list of carb ratios (g/U)

    sensitivity_starts -- list of time objects of start times of
                          given insulin sensitivity values
    sensitivity_ends -- list of time objects of start times of
                        given insulin sensitivity values
    sensitivity_values -- list of sensitivities (mg/dL/U)

    absorption_time_overrun -- multiplier to determine absorption time
                               from the specified absorption time
    default_absorption_time -- absorption time to use for unspecified
                               carb entries
    delay -- the time to delay the carb effect
    delta -- time interval between glucose values

    Output:
    3 lists in format (absorption_results, absorption_timelines, carb_entries)
        - lists are matched by index
            - one index represents one carb entry and its corresponding data

        - absorption_results: each index is a list of absorption information
            - structure: [(0) observed grams absorbed,
                          (1) clamped grams,
                          (2) total carbs in entry,
                          (3) remaining carbs,
                          (4) observed absorption start,
                          (5) observed absorption end,
                          (6) estimated time remaining]
        - absorption_timelines: each index is a list that contains
                                lists of timeline values
            - structure: [(0) timeline start time,
                          (1) timeline end time,
                          (2) absorbed value during timeline interval (g)]
            - if a timeline is a list with only "None", less than minimum
              absorption was observed
        - carb_entries: each index is a list of carb entry values
            - these lists are values that were calculated during map_ runtime
            - structure: [(0) carb sensitivities (mg/dL/G of carbohydrate),
                          (1) maximum carb absorption times (min),
                          (2) maximum absorption end times (datetime),
                          (3) last date effects were observed (datetime)
                          (4) total glucose effect expected for entry (mg/dL)]
    """
    assert len(carb_entry_starts) == len(carb_entry_quantities)\
        == len(carb_entry_absorptions), "expected input shapes to match"

    assert len(effect_starts) == len(effect_ends) == len(effect_values), \
        "expected input shapes to match"

    assert len(carb_ratio_starts) == len(carb_ratios),\
        "expected input shapes to match"

    assert len(sensitivity_starts) == len(sensitivity_ends)\
        == len(sensitivity_values), "expected input shapes to match"

    if (not carb_entry_starts
            or not carb_ratios
            or not sensitivity_starts):
        return ([], [], [])

    builder_entry_indexes = list(range(0, len(carb_entry_starts)))

    # CSF is in mg/dL/g
    builder_carb_sensitivities = [
        find_ratio_at_time(
            sensitivity_starts,
            sensitivity_ends,
            sensitivity_values,
            carb_entry_starts[i]
            ) /
        find_ratio_at_time(
            carb_ratio_starts,
            [],
            carb_ratios,
            carb_entry_starts[i]
            )
        for i in builder_entry_indexes
        ]

    # unit: g/s
    builder_max_absorb_times = [
        (carb_entry_absorptions[i]
         or default_absorption_time)
        * absorption_time_overrun
        for i in builder_entry_indexes
        ]

    builder_max_end_dates = [
        carb_entry_starts[i]
        + timedelta(minutes=builder_max_absorb_times[i] + delay)
        for i in builder_entry_indexes
        ]

    last_effect_dates = [
        min(
            builder_max_end_dates[i],
            max(
                effect_ends[len(effect_ends)-1],
                carb_entry_starts[i]
            )
        )
        for i in builder_entry_indexes
        ]

    entry_effects = [
        carb_entry_quantities[i] * builder_carb_sensitivities[i]
        for i in builder_entry_indexes
        ]

    observed_effects = [0 for i in builder_entry_indexes]
    observed_completion_dates = [None for i in builder_entry_indexes]

    observed_timeline_starts = [[] for i in builder_entry_indexes]
    observed_timeline_ends = [[] for i in builder_entry_indexes]
    observed_timeline_carb_values = [[] for i in builder_entry_indexes]

    assert len(builder_entry_indexes) == len(builder_carb_sensitivities)\
        == len(builder_max_absorb_times) == len(builder_max_end_dates)\
        == len(last_effect_dates), "expected shapes to match"

    def add_next_effect(entry_index, effect, start, end):

        if start < carb_entry_starts[entry_index]:
            return

        observed_effects[entry_index] += effect

        if not observed_completion_dates[entry_index]:
            # Continue recording the timeline until
            # 100% of the carbs have been observed
            observed_timeline_starts[entry_index].append(start)
            observed_timeline_ends[entry_index].append(end)
            observed_timeline_carb_values[entry_index].append(
                effect / builder_carb_sensitivities[entry_index]
            )

            # Once 100% of the carbs are observed, track the endDate
            if (
                    observed_effects[entry_index] + sys.float_info.epsilon
                    >= entry_effects[entry_index]
                ):
                observed_completion_dates[entry_index] = end

    for index in range(0, len(effect_starts)):

        if effect_starts[index] >= effect_ends[index]:
            continue

        # Select only the entries whose dates overlap the current date interval
        # These are not always contiguous, as maxEndDate varies between entries
        active_builders = []

        for j in builder_entry_indexes:
            if (effect_starts[index] < builder_max_end_dates[j]
                    and effect_starts[index] >= carb_entry_starts[j]):
                active_builders.append(j)

        # Ignore velocities < 0 when estimating carb absorption.
        # These are most likely the result of insulin absorption increases
        # such as during activity
        effect_value = max(0, effect_values[index]) * delta

        def rate_increase(index_):
            return (carb_entry_quantities[index_]
                    / builder_max_absorb_times[index_]
                    )
        # Sum the minimum absorption rates of each active entry to
        # determine how to split the active effects
        total_rate = 0

        for i in active_builders:
            total_rate += rate_increase(i)

        for b_index in active_builders:
            entry_effect = (carb_entry_quantities[b_index]
                            * builder_carb_sensitivities[b_index]
                           )
            remaining_effect = max(entry_effect - observed_effects[b_index], 0)

            # Apply a portion of the effect to this entry
            partial_effect_value = min(remaining_effect,
                                       (carb_entry_quantities[b_index]
                                        / builder_max_absorb_times[b_index]
                                        ) / total_rate * effect_value
                                       if total_rate != 0 and effect_value != 0
                                       else 0
                                       )

            total_rate -= (carb_entry_quantities[b_index]
                           / builder_max_absorb_times[b_index]
                           )
            effect_value -= partial_effect_value

            add_next_effect(
                b_index,
                partial_effect_value,
                effect_starts[index],
                effect_ends[index]
            )

            # If there's still remainder effects with no additional entries
            # to account them to, count them as overrun on the final entry
            if (effect_value > sys.float_info.epsilon
                    and b_index == (len(active_builders) - 1)
               ):
                add_next_effect(
                    b_index,
                    effect_value,
                    effect_starts[index],
                    effect_ends[index],
                    )

    def absorption_result(builder_index):
        # absorption list structure: [observed grams absorbed, clamped grams,
        # total carbs in entry, remaining carbs, observed absorption start,
        # observed absorption end, estimated time remaining]
        observed_grams = (observed_effects[builder_index]
                          / builder_carb_sensitivities[builder_index])

        entry_grams = carb_entry_quantities[builder_index]

        time = (time_interval_since(
            last_effect_dates[builder_index],
            carb_entry_starts[builder_index]
            ) / 60
                - delay
                )
        min_predicted_grams = linearly_absorbed_carbs(
            entry_grams,
            time,
            builder_max_absorb_times[builder_index]
        )
        clamped_grams = min(
            entry_grams,
            max(min_predicted_grams, observed_grams)
        )

        min_absorption_rate = (carb_entry_quantities[builder_index]
                               / builder_max_absorb_times[builder_index]
                               )
        estimated_time_remaining = ((entry_grams - clamped_grams)
                                    / min_absorption_rate
                                    if min_absorption_rate > 0
                                    else 0)

        absorption = [
            observed_grams,
            clamped_grams,
            entry_grams,
            entry_grams - clamped_grams,
            carb_entry_starts[builder_index],
            observed_completion_dates[builder_index]
            or last_effect_dates[builder_index],
            estimated_time_remaining
        ]

        return absorption

    # The timeline of observed absorption,
    # if greater than the minimum required absorption.
    def clamped_timeline(builder_index):
        entry_grams = carb_entry_quantities[builder_index]

        time = (time_interval_since(
            last_effect_dates[builder_index],
            carb_entry_starts[builder_index]
            ) / 60
                - delay
                )

        min_predicted_grams = linearly_absorbed_carbs(
            entry_grams,
            time,
            builder_max_absorb_times[builder_index]
        )
        observed_grams = (
            observed_effects[builder_index]
            / builder_carb_sensitivities[builder_index]
        )

        output = []
        for i in range(0, len(observed_timeline_starts[builder_index])):
            output.append(
                [
                    observed_timeline_starts[builder_index][i],
                    observed_timeline_ends[builder_index][i],
                    observed_timeline_carb_values[builder_index][i]
                ] if (
                    observed_grams >= min_predicted_grams
                    )
                else [None, None, None])

        return output

    def entry_properties(i):
        return [builder_carb_sensitivities[i],
                builder_max_absorb_times[i],
                builder_max_end_dates[i],
                last_effect_dates[i],
                entry_effects[i]
                ]

    entries = []
    absorptions = []
    timelines = []

    for i in builder_entry_indexes:
        absorptions.append(absorption_result(i))
        timelines.append(clamped_timeline(i))
        entries.append(entry_properties(i))

    assert len(absorptions) == len(timelines) == len(entries),\
        "expect output shapes to match"

    return (absorptions, timelines, entries)


def linearly_absorbed_carbs(total, time, absorption_time):
    """
    Find absorbed carbs using a linear model

    Parameters:
    total -- total grams of carbs
    time -- relative time after eating (in minutes)
    absorption_time --  time for carbs to completely absorb (in minutes)

    Output:
    Grams of absorbed carbs
    """
    return total * linear_percent_absorption_at_time(time, absorption_time)


def linear_unabsorbed_carbs(total, time, absorption_time):
    """
    Find unabsorbed carbs using a linear model

    Parameters:
    total -- total grams of carbs
    time -- relative time after eating (in minutes)
    absorption_time --  time for carbs to completely absorb (in minutes)

    Output:
    Grams of unabsorbed carbs
    """
    return total * (1 - linear_percent_absorption_at_time(time,
                                                          absorption_time
                                                          )
                    )


def linear_percent_absorption_at_time(time, absorption_time):
    """
    Find percent of absorbed carbs using a linear model

    Parameters:
    time -- relative time after eating (in minutes)
    absorption_time --  time for carbs to completely absorb (in minutes)

    Output:
    Percent of absorbed carbs
    """
    if time <= 0:
        return 0
    if time < absorption_time:
        return time / absorption_time
    return 1


def parabolic_absorbed_carbs(total, time, absorption_time):
    """
    Find absorbed carbs using a parabolic model

    Parameters:
    total -- total grams of carbs
    time -- relative time after eating (in minutes)
    absorption_time --  time for carbs to completely absorb (in minutes)

    Output:
    Grams of absorbed carbs
    """
    return total * parabolic_percent_absorption_at_time(time,
                                                        absorption_time
                                                        )


def parabolic_percent_absorption_at_time(time, absorption_time):
    """
    Find percent of absorbed carbs using a parabolic model

    Parameters:
    time -- relative time after eating (in minutes)
    absorption_time --  time for carbs to completely absorb (in minutes)

    Output:
    Percent of absorbed carbs
    """
    if time < 0:
        return 0

    if time <= absorption_time / 2:
        return 2 / pow(absorption_time, 2) * pow(time, 2)

    if time < absorption_time:
        return -1 + 4 / absorption_time * (time - pow(time, 2)
                                           / (2 * absorption_time)
                                           )
    return 1


def simulation_date_range(
        start_times,
        end_times,
        absorption_times,
        default_absorption_time,
        delay,
        delta,
        start=None,
        end=None,
        scaler=1
        ):
    """ Create date range based on carb data and user-specified parameters

    Arguments:
    start_times -- list of datetime object(s) of starts
    end_times -- list of datetime object(s) of ends
    absorption_times -- list of absorption times (in minutes)
    default_absorption_time -- length of interval
    delay -- additional time added to interval
    delta -- what to round to
    start -- specified start date
    end -- specified end date
    scaler -- the factor to extend the absorption time by

    Output:
    tuple with (start_time, end_time) structure
    """
    if not start_times:
        raise ValueError

    if start is not None and end is not None:
        return(date_floored_to_time_interval(start, delta),
               date_ceiled_to_time_interval(end, delta)
               )

    min_date = start_times[0]
    max_date = min_date

    for i in range(0, len(start_times)):
        if start_times[i] < min_date:
            min_date = start_times[i]

        try:
            end_date = end_times[i] + timedelta(
                (absorption_times or default_absorption_time) * scaler
                + delay
            )
        except IndexError:
            end_date = start_times[i] + timedelta(
                minutes=(
                    (absorption_times[i] or default_absorption_time) * scaler
                    + delay)
            )

        if end_date > max_date:
            max_date = end_date

    start_date = date_floored_to_time_interval(start or min_date, delta)
    end_date = date_ceiled_to_time_interval(end or max_date, delta)

    """ 
    Commented out due to a bug in carbs expiring. 
    //todo need to look into why carbs are not expiring. 
    assert start_date <= end_date,\
        "expected start to be less than or equal to end"
    """

    return (start_date, end_date)


def carbs_on_board(
        carb_starts, carb_quantities, carb_absorptions,
        default_absorption_time,
        delay=10,
        delta=5,
        start=None,
        end=None
        ):
    """
    Find the carbs on board non-dynamically

    Arguments:
    carb_starts -- list of times of carb entry (datetime objects)
    carb_quantities -- list of grams of carbs eaten
    carb_absorptions -- list of lengths of absorption times (mins)

    default_absorption_time -- absorption time to use for unspecified
                               carb entries
    delay -- the time to delay the carb effect
    delta -- the differential between timeline entries
    start -- datetime to start calculation of glucose effects
    end -- datetime to stop calculation of glucose effects

    Output:
    Two lists in format (carb on board start dates, carb on board values)
    """
    assert len(carb_starts) == len(carb_quantities)\
        == len(carb_absorptions), "expected input shapes to match"

    if not carb_starts:
        return ([], [])

    (start, end) = simulation_date_range(
        carb_starts,
        [],
        carb_absorptions,
        default_absorption_time,
        delay=delay,
        delta=delta,
        start=start,
        end=end
        )

    date = start
    cob_start_dates = []
    cob_values = []

    def find_partial_effect(i):
        return carbs_on_board_helper(
            carb_starts[i],
            carb_quantities[i],
            date,
            default_absorption_time,
            delay,
            carb_absorptions[i]
            )

    while date <= end:
        cob_sum = 0
        for i in range(0, len(carb_starts)):
            cob_sum += find_partial_effect(i)

        cob_start_dates.append(date)
        cob_values.append(cob_sum)
        date += timedelta(minutes=delta)

    assert len(cob_start_dates) == len(cob_values),\
        "expected output shapes to match"
    return (cob_start_dates, cob_values)


def carbs_on_board_helper(
        carb_start,
        carb_value,
        at_time,
        default_absorption_time,
        delay,
        carb_absorption_time=None
        ):
    """
    Find partial COB for a particular carb entry

    Arguments:
    carb_start -- time of carb entry (datetime objects)
    carb_value -- grams of carbs eaten

    at_date -- date to calculate the glucose effect (datetime object)

    default_absorption_time -- absorption time to use for unspecified
                               carb entries

    delay -- the time to delay the carb effect
    carb_absorption_time -- time carbs will take to absorb (mins)

    Output:
    Carbohydrate value (g)
    """
    time = time_interval_since(at_time, carb_start)
    delay *= 60

    if time >= 0:
        value = (carb_value
                 * (1 - parabolic_percent_absorption_at_time(
                     (time - delay) / 60,
                     carb_absorption_time or default_absorption_time
                     )
                    )
                 )
    else:
        value = 0
    return value


def dynamic_carbs_on_board(
        carb_starts, carb_quantities, carb_absorptions,
        absorptions, timelines,
        default_absorption_time,
        delay=10,
        delta=5,
        start=None,
        end=None,
        scaler=1
        ):
    """
    Find the carbs on board *dynamically*

    Arguments:
    carb_starts -- list of times of carb entry (datetime objects)
    carb_quantities -- list of grams of carbs eaten
    carb_absorptions -- list of lengths of absorption times (mins)

    absorptions -- list of lists of absorption information
                   (computed via map_)
    timelines -- list of lists of carb absorption timelines
                 (computed via map_)

    default_absorption_time -- absorption time to use for unspecified
                               carb entries
    delay -- the time to delay the carb effect
    delta -- the differential between timeline entries
    start -- datetime to start calculation of glucose effects
    end -- datetime to stop calculation of glucose effects
    scaler -- the factor to extend carb absorption by when calculating the
              end time for the COB timeline

    Output:
    Two lists in format (carb on board start dates, carb on board values)
    """
    assert len(carb_starts) == len(carb_quantities)\
        == len(carb_absorptions), "expected input shapes to match"

    assert len(absorptions) == len(timelines),\
        "expected input shapes to match"

    if not carb_starts or not absorptions:
        return ([], [])

    (start, end) = simulation_date_range(
        carb_starts,
        [],
        carb_absorptions,
        default_absorption_time,
        delay=delay,
        delta=delta,
        start=start,
        end=end,
        scaler=scaler
        )

    date = start
    cob_dates = []
    cob_values = []

    def find_partial_cob(i):
        return carb_status.dynamic_carbs_on_board_helper(
            carb_starts[i],
            carb_quantities[i],
            absorptions[i],
            timelines[i],
            date,
            default_absorption_time,
            delay,
            delta,
            carb_absorptions[i]
            )

    while date <= end:
        cob_sum = 0
        for i in range(0, len(carb_starts)):
            cob_sum += find_partial_cob(i)

        cob_dates.append(date)
        cob_values.append(cob_sum)
        date += timedelta(minutes=delta)

    assert len(cob_dates) == len(cob_values),\
        "expected output shapes to match"
    return (cob_dates, cob_values)


def carb_glucose_effects(
        carb_starts, carb_quantities, carb_absorptions,
        carb_ratio_starts, carb_ratios,
        sensitivity_starts, sensitivity_ends, sensitivity_values,
        default_absorption_time,
        delay=10,
        delta=5,
        start=None,
        end=None,
        scaler=1
        ):
    """
    Find the expected effects of carbohydate consumption on blood glucose

    Arguments:
    carb_starts -- list of times of carb entry (datetime objects)
    carb_quantities -- list of grams of carbs eaten
    carb_absorptions -- list of lengths of absorption times (mins)

    carb_ratio_starts -- list of start times of carb ratios (time objects)
    carb_ratios -- list of carb ratios (g/U)

    sensitivity_starts -- list of time objects of start times of
                          given insulin sensitivity values
    sensitivity_ends -- list of time objects of start times of
                        given insulin sensitivity values
    sensitivity_values -- list of sensitivities (mg/dL/U)

    default_absorption_time -- absorption time to use for unspecified
                               carb entries
    delay -- the time to delay the carb effect
    delta -- the differential between timeline entries
    start -- datetime to start calculation of glucose effects
    end -- datetime to stop calculation of glucose effects
    scaler -- the factor to extend carb absorption by when calculating the
              end time for the COB timeline

    Output:
    Two lists in format (effect_start_dates, effect_values)
    """
    assert len(carb_starts) == len(carb_quantities)\
        == len(carb_absorptions), "expected input shapes to match"

    assert len(carb_ratio_starts) == len(carb_ratios),\
        "expected input shapes to match"

    assert len(sensitivity_starts) == len(sensitivity_ends)\
        == len(sensitivity_values), "expected input shapes to match"

    if (not carb_starts
            or not carb_ratio_starts
            or not sensitivity_starts
       ):
        return ([], [])

    (start, end) = simulation_date_range(
        carb_starts,
        [],
        carb_absorptions,
        default_absorption_time,
        delay=delay,
        delta=delta,
        start=start,
        end=end,
        scaler=scaler
        )

    date = start
    effect_start_dates = []
    effect_values = []

    def find_partial_effect(i):
        insulin_sensitivity = find_ratio_at_time(
            sensitivity_starts,
            sensitivity_ends,
            sensitivity_values,
            carb_starts[i]
            )
        carb_ratio = find_ratio_at_time(
            carb_ratio_starts,
            [],
            carb_ratios,
            carb_starts[i]
            )
        return carb_glucose_effect(
            carb_starts[i],
            carb_quantities[i],
            date,
            carb_ratio,
            insulin_sensitivity,
            default_absorption_time,
            delay,
            carb_absorptions[i]
            )

    while date <= end:
        effect_sum = 0
        for i in range(0, len(carb_starts)):
            effect_sum += find_partial_effect(i)

        effect_start_dates.append(date)
        effect_values.append(effect_sum)
        date += timedelta(minutes=delta)

    assert len(effect_start_dates) == len(effect_values),\
        "expected output shapes to match"
    return (effect_start_dates, effect_values)


def carb_glucose_effect(
        carb_start, carb_value,
        at_date,
        carb_ratio,
        insulin_sensitivity,
        default_absorption_time,
        delay,
        carb_absorption_time=None
        ):
    """
    Find partial effect of carbohydate consumption on blood glucose

    Arguments:
    carb_start -- time of carb entry (datetime objects)
    carb_value -- grams of carbs eaten

    at_date -- date to calculate the glucose effect (datetime object)

    carb_ratio -- ratio (g of carbs/U)
    insulin_sensitivity -- sensitivity (mg/dL/U)

    default_absorption_time -- absorption time to use for unspecified
                               carb entries

    delay -- the time to delay the carb effect
    carb_absorption_time -- time carbs will take to absorb (mins)

    Output:
    Glucose effect (mg/dL/min)
    """

    return insulin_sensitivity / carb_ratio * absorbed_carbs(
        carb_start,
        carb_value,
        carb_absorption_time or default_absorption_time,
        at_date,
        delay
        )


def dynamic_glucose_effects(
        carb_starts, carb_quantities, carb_absorptions,
        absorptions, timelines,
        carb_ratio_starts, carb_ratios,
        sensitivity_starts, sensitivity_ends, sensitivity_values,
        default_absorption_time,
        delay=10,
        delta=5,
        start=None,
        end=None,
        scaler=1
        ):
    """
    Find the expected effects of carbohydate consumption on blood glucose
    dynamically

    Arguments:
    carb_starts -- list of times of carb entry (datetime objects)
    carb_quantities -- list of grams of carbs eaten
    carb_absorptions -- list of lengths of absorption times (mins)

    absorptions -- list of lists of absorption information
                   (computed via map_)
    timelines -- list of lists of carb absorption timelines
                 (computed via map_)

    carb_ratio_starts -- list of start times of carb ratios (time objects)
    carb_ratios -- list of carb ratios (g/U)


    sensitivity_starts -- list of time objects of start times of
                          given insulin sensitivity values
    sensitivity_ends -- list of time objects of start times of
                        given insulin sensitivity values
    sensitivity_values -- list of sensitivities (mg/dL/U)

    default_absorption_time -- absorption time to use for unspecified
                               carb entries
    delay -- the time to delay the carb effect
    delta -- the differential between timeline entries
    start -- datetime to start calculation of glucose effects
    end -- datetime to stop calculation of glucose effects
    scaler -- the factor to extend carb absorption by when calculating the
              end time for the COB timeline

    Output:
    Two lists in format (effect_start_dates, effect_values)
    """
    assert len(carb_starts) == len(carb_quantities)\
        == len(carb_absorptions), "expected input shapes to match"

    assert len(carb_ratio_starts) == len(carb_ratios),\
        "expected input shapes to match"

    assert len(sensitivity_starts) == len(sensitivity_ends)\
        == len(sensitivity_values), "expected input shapes to match"

    assert len(absorptions) == len(timelines),\
        "expected input shapes to match"

    if (not carb_starts
            or not carb_ratio_starts
            or not sensitivity_starts
            or not absorptions
       ):
        return ([], [])

    (start, end) = simulation_date_range(
        carb_starts,
        [],
        carb_absorptions,
        default_absorption_time,
        delay=delay,
        delta=delta,
        start=start,
        end=end,
        scaler=scaler
        )

    date = start
    effect_start_dates = []
    effect_values = []

    def find_partial_effect(i):
        insulin_sensitivity = find_ratio_at_time(
            sensitivity_starts,
            sensitivity_ends,
            sensitivity_values,
            carb_starts[i]
            )
        carb_ratio = find_ratio_at_time(
            carb_ratio_starts,
            [],
            carb_ratios,
            carb_starts[i]
            )
        csf = insulin_sensitivity / carb_ratio
        partial_carbs_absorbed = carb_status.dynamic_absorbed_carbs(
            carb_starts[i],
            carb_quantities[i],
            absorptions[i],
            timelines[i],
            date,
            carb_absorptions[i] or default_absorption_time,
            delay,
            delta,
        )

        return csf * partial_carbs_absorbed

    while date <= end:
        effect_sum = 0
        for i in range(0, len(carb_starts)):
            effect_sum += find_partial_effect(i)

        effect_start_dates.append(date)
        effect_values.append(effect_sum)
        date += timedelta(minutes=delta)

    assert len(effect_start_dates) == len(effect_values),\
        "expected output shapes to match"
    return (effect_start_dates, effect_values)


def absorbed_carbs(
        start_date,
        carb_value,
        absorption_time,
        at_date,
        delay
        ):
    """
    Find absorbed carbs using a parabolic model

    Parameters:
    start_date -- date of carb consumption (datetime object)
    carb_value -- carbs consumed
    absorption_time --  time for carbs to completely absorb (in minutes)
    at_date -- date to calculate the absorbed carbs (datetime object)
    delay -- minutes to delay the start of absorption

    Output:
    Grams of absorbed carbs
    """
    time = time_interval_since(at_date, start_date) / 60

    return parabolic_absorbed_carbs(
        carb_value,
        time - delay,
        absorption_time
        )


def filter_date_range_for_carbs(
        starts, values, absorptions,
        start_date,
        end_date
        ):
    """ Returns an array of elements filtered by the specified date range.

    Arguments:
    starts -- start dates (datetime)
    values -- carb values (g)
    absorptions -- absorption times for entries

    start_date -- the earliest date of elements to return
    end_date -- the last date of elements to return

    Output:
    Filtered carb entries in format (starts, values, absorptions)
    """
    # ends might not necesarily be the same length as starts/values
    # because not all types have "end dates"
    assert len(starts) == len(values) == len(absorptions),\
        "expected input shapes to match"

    (filtered_starts,
     filtered_values,
     filtered_absorptions
     ) = ([], [], [])

    for i in range(0, len(starts)):
        if start_date and starts[i] < start_date:
            continue

        if end_date and starts[i] > end_date:
            continue

        filtered_starts.append(starts[i])
        filtered_values.append(values[i])
        filtered_absorptions.append(absorptions[i])

    assert len(filtered_starts) == len(filtered_values)\
        == len(filtered_absorptions), "expected output shapes to match"

    return (
        filtered_starts,
        filtered_values,
        filtered_absorptions
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 15 09:54:54 2019

@author: annaquinlan
"""
from enum import Enum


class DoseType(Enum):
    suspend = 0
    resume = 1
    basal = 2
    tempbasal = 3
    bolus = 4
    meal = 5  # meals are included for compatability with Loop tests

    @staticmethod
    def from_str(label):
        if label.lower() in ["suspend", "pumpsuspend"]:
            return DoseType.suspend
        elif label.lower() in ["resume", "pumpresume"]:
            return DoseType.resume
        elif label.lower() in ["basal", "basalprofilestart"]:
            return DoseType.basal
        elif label.lower() == "tempbasal":
            return DoseType.tempbasal
        elif label.lower() == "bolus":
            return DoseType.bolus
        elif label.lower() == "meal":
            return DoseType.meal
        else:
            raise NotImplementedError(label, "not recognized")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul 10 17:03:59 2019

@author: annaquinlan

Github URL: https://github.com/tidepool-org/LoopKit/blob/
57a9f2ba65ae3765ef7baafe66b883e654e08391/LoopKit/InsulinKit/DoseStore.swift
"""
# pylint: disable=R0913, R0914, C0200
from datetime import timedelta

from pyloopkit.dose_math import filter_date_range_for_doses
from pyloopkit.insulin_math import (annotated, trim, glucose_effects, reconciled)
from pyloopkit.loop_math import filter_date_range, sort_dose_lists


def get_glucose_effects(
        types, starts, ends, values, delivered_units,
        start_date,
        basal_starts, basal_rates, basal_minutes,
        sensitivity_starts, sensitivity_ends, sensitivity_values,
        insulin_model,
        delay=10,
        end_date=None
        ):
    """ Get the glucose effects at a particular time, given a list of
    doses and a time interval

    Arguments:
    types -- list of types of dose (basal, bolus, etc)
    starts -- start dates of the doses (datetime obj)
    ends -- end dates of the doses (datetime obj)
    values -- actual basal rates of doses in U/hr (if a basal)
             or the value of the boluses if in U
    delivered_units -- net Units of insulin actually delivered by a dose

    start_date -- date to start calculating glucose effects

    basal_starts -- list of times the basal rates start at
    basal_rates -- list of basal rates(U/hr)
    basal_minutes -- list of basal lengths (in mins)

    sensitivity_starts -- list of time objects of start times of
                          given insulin sensitivity values
    sensitivity_ends -- list of time objects of start times of
                        given insulin sensitivity values
    sensitivity_values -- list of sensitivities (mg/dL/U)

    insulin_model -- list in format [DIA (in hours)] if Walsh model, or
                     [DIA (minutes), peak (minutes)] if exponential model

    end_date -- date to stop calculating glucose effects

    Output:
    Glucose effects in the format (effect_date, effect_value)
    """
    assert len(types) == len(starts) == len(ends) == len(values) == len(delivered_units),\
        "expected input shapes to match"

    # to properly know glucose effects at start_date,
    # we need to go back another DIA hours
    if len(insulin_model) == 1:  # if using Walsh model
        dose_start = (start_date
                      - timedelta(
                          hours=insulin_model[0]
                          )
                      )
    else:
        dose_start = (start_date
                      - timedelta(
                          minutes=insulin_model[0]
                          )
                      )

    filtered_doses = filter_date_range_for_doses(
        types, starts, ends, values, delivered_units,
        dose_start,
        end_date
        )

    # reconcile the doses to get a cleaner data set
    # (add resumes for suspends and trim any overlapping temp basals)
    reconciled_doses = reconciled(
        *filtered_doses
        )
    # sort the lists because they could be slightly out of order due to
    # basals and suspends
    sorted_reconciled_doses = sort_dose_lists(*reconciled_doses)[0:5]

    # annotate the doses with scheduled basal rate
    (a_types,
     a_starts,
     a_ends,
     a_values,
     a_scheduled_rates,
     a_delivered_units
     ) = annotated(
         *sorted_reconciled_doses,
         basal_starts, basal_rates, basal_minutes,
         convert_to_units_hr=False
     )

    # trim the doses to start of interval
    for i in range(0, len(a_types)):
        result = trim(
            a_types[i], a_starts[i], a_ends[i], a_values[i], a_delivered_units[i],
            a_scheduled_rates[i],
            start_interval=dose_start
            )

        a_starts[i] = result[1]
        a_ends[i] = result[2]

    # get the glucose effects using the prepared dose data
    glucose_effect = glucose_effects(
        a_types, a_starts, a_ends, a_values, a_scheduled_rates, a_delivered_units,
        insulin_model,
        sensitivity_starts, sensitivity_ends, sensitivity_values,
        delay=delay,
        start=start_date,
        end=end_date
        )

    # don't return effects that are less than the start date or greater than
    # the end date (if there is one)
    (filtered_starts,
     ends,
     filtered_effect_values) = filter_date_range(
         glucose_effect[0],
         [],
         glucose_effect[1],
         start_date,
         end_date
         )

    return (filtered_starts, filtered_effect_values)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul 16 15:31:31 2019

@author: annaquinlan

Github URL: https://github.com/tidepool-org/LoopKit/blob/
57a9f2ba65ae3765ef7baafe66b883e654e08391/LoopKit/CarbKit/CarbStore.swift
"""
# pylint: disable=R0913, R0914
from datetime import timedelta

from pyloopkit.carb_math import (filter_date_range_for_carbs, map_, carb_glucose_effects,
                       dynamic_glucose_effects, dynamic_carbs_on_board,
                       carbs_on_board)


def get_carb_glucose_effects(
        carb_dates, carb_values, absorption_times,
        at_date,
        effect_starts, effect_ends, effect_values,
        carb_ratio_starts, carb_ratios,
        sensitivity_starts, sensitivity_ends, sensitivity_values,
        default_absorption_times,
        absorption_time_overrun=1.5,
        delay=10,
        delta=5,
        end_date=None
        ):
    """ Retrieve a timeline of effect on blood glucose from carbohydrates

    Arguments:
    carb_dates -- list of times of carb entry (datetime objects)
    carb_values -- list of grams of carbs eaten
    absorption_times -- list of lengths of absorption times (mins)

    at_date -- the time to calculate the effect at (datetime object)

    effect_starts -- list of start times of carb effect (datetime objects)
    effect_ends -- list of end times of carb effect (datetime objects)
    effect_values -- list of glucose velocities (mg/dL)

    carb_ratio_starts -- list of start times of carb ratios (time objects)
    carb_ratios -- list of carb ratios (g/U)

    sensitivity_starts -- list of time objects of start times of
                          given insulin sensitivity values
    sensitivity_ends -- list of time objects of start times of
                        given insulin sensitivity values
    sensitivity_values -- list of sensitivities (mg/dL/U)

    default_absorption_time -- absorption time to use for unspecified
                               carb entries

    absorption_time_overrun -- multiplier to determine absorption time
                               from the specified absorption time

    delay -- the time to delay the carb effect
    delta -- time interval between glucose values

    end_date -- date to end calculation of glucose effects

    Output:
    An array of effects in chronological order
    """
    assert len(carb_dates) == len(carb_values) == len(absorption_times),\
        "expected input shapes to match"

    assert len(effect_starts) == len(effect_ends) == len(effect_values),\
        "expected input shapes to match"

    if not carb_dates:
        return ([], [])

    maximum_absorption_time_interval = default_absorption_times[2] * 2

    # To know glucose effects at the requested start date, we need to fetch
    # samples that might still be absorbing
    food_start = at_date - timedelta(minutes=maximum_absorption_time_interval)

    filtered_carbs = filter_date_range_for_carbs(
        carb_dates, carb_values, absorption_times,
        food_start,
        end_date
        )

    # if we have counteraction effects, generate our carb glucose effects
    # with a dynamic model
    if effect_starts and effect_starts[0]:
        (absorption_results,
         timelines
         ) = map_(
             *filtered_carbs,
             effect_starts, effect_ends, effect_values,
             carb_ratio_starts, carb_ratios,
             sensitivity_starts, sensitivity_ends, sensitivity_values,
             absorption_time_overrun,
             default_absorption_times[1],
             delay,
             delta
             )[0:2]

        effects = dynamic_glucose_effects(
            *filtered_carbs,
            absorption_results, timelines,
            carb_ratio_starts, carb_ratios,
            sensitivity_starts, sensitivity_ends, sensitivity_values,
            default_absorption_times[1],
            delay,
            delta,
            start=at_date,
            end=end_date,
            scaler=1.7
            )
    # otherwise, use a static model
    else:
        effects = carb_glucose_effects(
            *filtered_carbs,
            carb_ratio_starts, carb_ratios,
            sensitivity_starts, sensitivity_ends, sensitivity_values,
            default_absorption_times[1],
            delay,
            delta,
            at_date,
            end_date
            )

    assert len(effects[0]) == len(effects[1]), "expected output shape to match"

    return effects


def get_carbs_on_board(
        carb_dates, carb_values, absorption_times,
        at_date,
        effect_starts, effect_ends, effect_values,
        carb_ratio_starts, carb_ratios,
        sensitivity_starts, sensitivity_ends, sensitivity_values,
        default_absorption_times,
        absorption_time_overrun=1.5,
        delay=10,
        delta=5,
        end_date=None
        ):
    """ Retrieves the COB at a time, or a timeline of COB

    Arguments:
    carb_dates -- list of times of carb entry (datetime objects)
    carb_values -- list of grams of carbs eaten
    absorption_times -- list of lengths of absorption times (mins)

    at_date -- the time to calculate the COB at (datetime object)

    effect_starts -- list of start times of carb effect (datetime objects)
    effect_ends -- list of end times of carb effect (datetime objects)
    effect_values -- list of glucose velocities (mg/dL)

    carb_ratio_starts -- list of start times of carb ratios (time objects)
    carb_ratios -- list of carb ratios (g/U)

    sensitivity_starts -- list of time objects of start times of
                          given insulin sensitivity values
    sensitivity_ends -- list of time objects of start times of
                        given insulin sensitivity values
    sensitivity_values -- list of sensitivities (mg/dL/U)

    default_absorption_times -- list absorption times to use for unspecified
                               carb entries in format [fast, medium, slow]

    absorption_time_overrun -- multiplier to determine absorption time
                               from the specified absorption time

    delay -- the time to delay the COB
    delta -- time interval between glucose values

    end_date -- date to end calculation of COB

    Output:
    COB timeline
    """
    assert len(carb_dates) == len(carb_values) == len(absorption_times),\
        "expected input shapes to match"

    assert len(effect_starts) == len(effect_ends) == len(effect_values),\
        "expected input shapes to match"

    if not carb_dates:
        return ([], [])

    start_date = at_date - timedelta(minutes=delta)

    maximum_absorption_time_interval = default_absorption_times[2] * 2

    # To know COB at the requested start date, we need to fetch samples that
    # might still be absorbing
    food_start = start_date - timedelta(
        minutes=maximum_absorption_time_interval
    )

    filtered_carbs = filter_date_range_for_carbs(
        carb_dates, carb_values, absorption_times,
        food_start,
        end_date
        )

    # If we have counteraction effects, use a dynamic model
    if (effect_starts and effect_starts[0]
            and carb_ratio_starts
            and sensitivity_starts
       ):
        (absorption_results,
         timelines
         ) = map_(
             *filtered_carbs,
             effect_starts, effect_ends, effect_values,
             carb_ratio_starts, carb_ratios,
             sensitivity_starts, sensitivity_ends, sensitivity_values,
             absorption_time_overrun,
             default_absorption_times[1],
             delay,
             delta
             )[0:2]

        cob_data = dynamic_carbs_on_board(
            *filtered_carbs,
            absorption_results, timelines,
            default_absorption_times[1],
            delay,
            delta,
            start=start_date,
            end=end_date,
            scaler=1.5
            )
    # otherwise, use a static model
    else:
        cob_data = carbs_on_board(
            *filtered_carbs,
            default_absorption_times[1],
            delay,
            delta,
            start=start_date,
            end=end_date
            )

    return cob_data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 20 16:04:11 2019

@author: annaquinlan

https://github.com/tidepool-org/LoopKit/blob/
57a9f2ba65ae3765ef7baafe66b883e654e08391/LoopKit/InsulinKit/
ExponentialInsulinModel.swift
"""
# pylint: disable=C0103
import math


def percent_effect_remaining(time, action_duration, peak_activity_time):
    """ Returns the percentage of total insulin effect remaining at a specified
        interval after delivery; aka insulin on board (IOB)

        This model allows us to specify time of peak activity, as well as
        duration, and provides activity and IOB decay functions

    Arguments:
    time -- the minutes after insulin delivery (it can be negative)
    action_duration -- the total duration on insulin activity (DIA)
    peak_activity_time -- the time (in minutes) of the peak of insulin activity
                          from dose

    Output:
    The percentage of total insulin effect remaining
    """

    if time <= 0:
        return 1
    if time > action_duration:
        return 0

    tau = (peak_activity_time * (1 - peak_activity_time / action_duration) /
           (1 - 2 * peak_activity_time / action_duration)
           )
    a = 2 * tau / action_duration
    S = 1 / (1 - a + (1 + a) * math.exp(-action_duration / tau))

    return 1 - S * (1 - a) * ((pow(time, 2) / (tau * action_duration * (1 - a))
                               - time / tau - 1) * math.exp(-time / tau) + 1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 14 10:40:48 2019

@author: annaquinlan

Github URL: https://github.com/tidepool-org/LoopKit/blob/
            57a9f2ba65ae3765ef7baafe66b883e654e08391/LoopKit/LoopMath.swift
"""
# pylint: disable=R0913, R0914, C0200, R0912, R0915, W0102, C0103
# disable pylint errors for too many arguments/variables
from datetime import timedelta
import numpy

from pyloopkit.date import (
    date_floored_to_time_interval,
    date_ceiled_to_time_interval,
    time_interval_since,
)


def predict_glucose(
    starting_date,
    starting_glucose,
    momentum_dates=[],
    momentum_values=None,
    carb_effect_dates=[],
    carb_effect_values=None,
    insulin_effect_dates=[],
    insulin_effect_values=None,
    correction_effect_dates=[],
    correction_effect_values=None,
):
    """Calculates a timeline of predicted glucose values
        from a variety of effects timelines.

    Each effect timeline:
     - Is given equal weight (exception: momentum effect timeline)
     - Can be of arbitrary size and start date
     - Should be in ascending order
     - Should have aligning dates with any overlapping timelines to ensure
       a smooth result

    Parameters:
    starting_date -- time of starting_glucose (datetime object)
    starting_glucose -- glucose value to use in predictions

    momentum_dates -- times of calculated momentums (datetime)
    momentum_values -- values (mg/dL) of momentums

    carb_effect_dates -- times of carb effects (datetime)
    carb_effect -- values (mg/dL) of effects from carbs

    insulin_effect_dates -- times of insulin effects (datetime)
    insulin_effect -- values (mg/dL) of effects from insulin

    correction_effect_dates -- times of retrospective effects (datetime)
    correction_effect -- values (mg/dL) retrospective glucose effects

    Output:
    Glucose predictions in form (prediction_times, prediction_glucose_values)
    """
    if momentum_dates:
        assert len(momentum_dates) == len(
            momentum_values
        ), "expected input shapes to match"

    if carb_effect_dates:
        assert len(carb_effect_dates) == len(
            carb_effect_values
        ), "expected input shapes to match"

    if insulin_effect_dates:
        assert len(insulin_effect_dates) == len(
            insulin_effect_values
        ), "expected input shapes to match"

    if correction_effect_dates:
        assert len(correction_effect_dates) == len(
            correction_effect_values
        ), "expected input shapes to match"

    # if we didn't get any effect data, we won't predict the glucose
    if (
        not momentum_dates
        and not carb_effect_dates
        and not insulin_effect_dates
        and not correction_effect_dates
    ):
        return ([], [])

    merged_dates = sorted(
        list(
            dict.fromkeys(
                momentum_dates
                + carb_effect_dates
                + insulin_effect_dates
                + correction_effect_dates
            )
        )
    )

    merged_values = [0 for i in merged_dates]

    if carb_effect_dates:
        previous_effect_value = carb_effect_values[0] or 0
        for i in range(0, len(carb_effect_dates)):
            value = carb_effect_values[i]
            list_index = merged_dates.index(carb_effect_dates[i])
            merged_values[list_index] = value - previous_effect_value

            previous_effect_value = value

    if insulin_effect_dates:
        previous_effect_value = insulin_effect_values[0] or 0
        for i in range(0, len(insulin_effect_dates)):
            value = insulin_effect_values[i]
            list_index = merged_dates.index(insulin_effect_dates[i])
            merged_values[list_index] = (
                merged_values[list_index] + value - previous_effect_value
            )

            previous_effect_value = value

    if correction_effect_dates:
        previous_effect_value = correction_effect_values[0] or 0
        for i in range(0, len(correction_effect_dates)):
            value = correction_effect_values[i]
            list_index = merged_dates.index(correction_effect_dates[i])
            merged_values[list_index] = (
                merged_values[list_index] + value - previous_effect_value
            )

            previous_effect_value = value

    # Blend the momentum effect linearly into the summed effect list
    if len(momentum_dates) > 1:
        previous_effect_value = momentum_values[0]

        # The blend begins delta minutes after after the last glucose (1.0)
        # and ends at the last momentum point (0.0)
        # This assumes the first one occurs on/before the starting glucose
        blend_count = len(momentum_dates) - 2
        time_delta = time_interval_since(momentum_dates[1], momentum_dates[0])
        # The difference between the first momentum value
        # and the starting glucose value
        momentum_offset = time_interval_since(starting_date, momentum_dates[0])

        blend_slope = 1 / blend_count
        blend_offset = momentum_offset / time_delta * blend_slope

        for i in range(0, len(momentum_dates)):
            value = momentum_values[i]
            date = momentum_dates[i]
            merge_index = merged_dates.index(date)

            effect_value_change = value - previous_effect_value

            split = min(
                1,
                max(
                    0,
                    (len(momentum_dates) - i) / blend_count
                    - blend_slope
                    + blend_offset,
                ),
            )
            effect_blend = (1 - split) * merged_values[merge_index]
            momentum_blend = split * effect_value_change

            merged_values[merge_index] = effect_blend + momentum_blend
            previous_effect_value = value

    predicted_dates = [starting_date]
    predicted_values = [starting_glucose]

    for i in range(0, len(merged_dates)):
        if merged_dates[i] > starting_date:
            last_value = predicted_values[-1]

            predicted_dates.append(merged_dates[i])
            predicted_values.append(last_value + merged_values[i])

    assert len(predicted_dates) == len(
        predicted_values
    ), "expected output shapes to match"
    return (predicted_dates, predicted_values)


def decay_effect(glucose_date, glucose_value, rate, duration, delta=5):
    """Calculates a timeline of glucose effects by applying a
        linear decay to a rate of change.

    Arguments:
    glucose_date -- time of glucose value (datetime)
    glucose_value -- value at the time of glucose_date
    rate -- the glucose velocity
    duration -- the duration the effect should continue before ending
    delta -- the time differential for the returned values

    Output:
    Glucose effects in format (effect_date, effect_value)
    """
    (start_date, end_date) = simulation_date_range_for_samples(
        [glucose_date], [], duration, delta
    )

    # The starting rate, which we will decay to 0 over the specified duration
    intercept = rate
    last_value = glucose_value
    effect_dates = [start_date]
    effect_values = [glucose_value]

    date = decay_start_date = start_date + timedelta(minutes=delta)
    slope = -intercept / (duration - delta)

    while date < end_date:
        value = (
            last_value
            + (intercept + slope * time_interval_since(date, decay_start_date) / 60)
            * delta
        )

        effect_dates.append(date)
        effect_values.append(value)

        last_value = value
        date = date + timedelta(minutes=delta)

    assert len(effect_dates) == len(effect_values), "expected output shapes to match"

    return (effect_dates, effect_values)


def simulation_date_range_for_samples(
    start_times, end_times, duration, delta, start=None, end=None, delay=0
):
    """Create date range based on samples and user-specified parameters

    Arguments:
    start_times -- list of datetime object(s) at start
    end_times -- list of datetime object(s) at end
    duration -- length of interval
    delta -- what to round to
    start -- specified start date
    end -- specified end date
    delay -- additional time added to interval in minutes

    Output:
    tuple with (start_time, end_time) structure
    """
    if start is not None and end is not None:
        return (
            date_floored_to_time_interval(start, delta),
            date_ceiled_to_time_interval(end, delta),
        )

    if not start_times:
        raise ValueError

    min_date = start_times[0]
    max_date = min_date

    for i in range(0, len(start_times)):
        if start_times[i] < min_date:
            min_date = start_times[i]

        try:
            if end_times[i] > max_date:
                max_date = end_times[i]
        # if end_times is an empty list, don't error
        except IndexError:
            continue

    start_date = date_floored_to_time_interval(start or min_date, delta)
    end_date = date_ceiled_to_time_interval(
        end or max_date + timedelta(minutes=duration + delay), delta
    )

    assert start_date <= end_date, "expected start to be less than end"

    return (start_date, end_date)


def subtracting(
    starts, ends, values, other_starts, other_ends, other_values, effect_interval
):
    """Subtracts an array of glucose effects with uniform intervals and
        no gaps from the collection of effect changes, which may not
        have uniform intervals.

    Parameters:
    starts -- start times of effect that is subtracted-from (datetime)
    ends -- end times of effect that is subtracted-from (datetime)
    values -- values of effect that is subtracted-from (datetime)

    other_starts -- start times of the effect to subtract (datetime)
    other_ends -- end times of the effect to subtract (datetime) (optional)
    other_values -- values of the effect to subtract

    effect_interval -- time interval (in minutes) between times in the
                       other_starts and other_ends lists

    Output:
    The resulting glucose effects in the form
    (start_times, end_times, values)
    """
    assert len(starts) == len(ends) == len(values), "expected input shapes to match"
    assert len(other_starts) == len(other_values), "expected input shapes to match"
    # Trim both collections to match
    (other_starts, other_ends, other_values) = filter_date_range(
        other_starts, other_ends, other_values, ends[0], None
    )
    (starts, ends, values) = filter_date_range(
        starts, ends, values, other_starts[0], None
    )

    (subtracted_starts, subtracted_values) = ([], [])

    previous_other_effect_value = other_values[0]
    effect_index = 0

    for i in range(1, len(other_starts)):
        if effect_index >= len(starts):
            break

        other_effect_value = other_values[i]
        other_effect_change = other_effect_value - previous_other_effect_value

        previous_other_effect_value = other_effect_value

        # Our effect array may have gaps, or have longer segments than 5 mins
        if other_ends[i] and ends[effect_index] > other_ends[i]:
            continue  # move on to next one

        if other_starts[i] and ends[effect_index] > other_starts[i]:
            continue  # move on to next one

        effect_value = values[effect_index]
        effect_value_matching_other_effect_interval = effect_value * effect_interval

        subtracted_starts.append(ends[effect_index])
        subtracted_values.append(
            effect_value_matching_other_effect_interval - other_effect_change
        )

        effect_index += 1

    # if we have run out of other_effect items,
    # we assume the other_effect_change remains zero
    for i in range(effect_index, len(starts)):
        effect_value = values[i]
        effect_value_matching_other_effect_interval = effect_value * effect_interval

        subtracted_starts.append(ends[i])
        subtracted_values.append(effect_value_matching_other_effect_interval)

    assert len(subtracted_starts) == len(
        subtracted_values
    ), "expected output shapes to match"

    return (subtracted_starts, subtracted_values)


def filter_date_range(starts, ends, values, start_date, end_date):
    """Returns tuple of elements filtered by the specified date range.

    Arguments:
    starts -- start dates (datetime)
    ends -- end dates (datetime)
    values -- glucose values

    start_date -- the earliest date of elements to return
    end_date -- the last date of elements to return

    Output:
    Filtered dates in format (starts, ends, values)
    """
    # ends might not necesarily be the same length as starts/values
    # because not all types have "end dates"
    assert len(starts) == len(values), "expected input shapes to match"

    (filtered_starts, filtered_ends, filtered_values) = ([], [], [])

    for i in range(0, len(starts)):
        if start_date and ends and ends[i] < start_date:
            continue

        if start_date and not ends and starts[i] < start_date:
            continue

        if end_date and starts[i] > end_date:
            continue

        filtered_starts.append(starts[i])
        filtered_ends.append(ends[i] if ends else None)
        filtered_values.append(values[i])

    assert (
        len(filtered_starts) == len(filtered_ends) == len(filtered_values)
    ), "expected output shapes to match"

    return (filtered_starts, filtered_ends, filtered_values)


def sort_dose_lists(list_1, list_2, list_3=None, list_4=None, list_5=None):
    """Sort dose lists that are matched index-wise, using the *second* list as
         the property to sort by

    Example:
        list_1: [50, 2, 3]               ->     [2, 3, 50]
        list_2: [dog, cat, parrot]       ->     [cat, parrot, dog]
    """
    unsort_1 = numpy.array(list_1)
    unsort_2 = numpy.array(list_2)
    unsort_3 = numpy.array(list_3)
    unsort_4 = numpy.array(list_4)
    unsort_5 = numpy.array(list_5)

    sort_indexes = unsort_2.argsort()

    l1 = list(unsort_1[sort_indexes])
    unsort_2.sort()
    list_2 = list(unsort_2)
    if list_3:
        l3 = list(unsort_3[sort_indexes])
    else:
        l3 = []
    if list_4:
        l4 = list(unsort_4[sort_indexes])
    else:
        l4 = []
    if list_5:
        l5 = list(unsort_5[sort_indexes])
    else:
        l5 = []

    return (l1, list_2, l3, l4, l5)


def combined_sums(starts, ends, values, duration):
    """
    Sums adjacent glucose effects into buckets of the specified duration.

    Requires the receiver to be sorted chronologically by endDate

    Arguments:
    starts -- start dates (datetime)
    ends -- end dates (datetime)
    values -- glucose values

    duration -- duration of each resulting summed element (minutes)

    Output:
    Summed effects in format (start_dates, end_dates, values)
    """
    assert len(starts) == len(values), "expected input shapes to match"

    sum_starts = []
    sum_ends = []
    sum_values = []

    starts = list(reversed(starts))
    ends = list(reversed(ends))
    values = list(reversed(values))
    last_valid_index = 0

    for i in range(0, len(starts)):
        sum_starts.append(starts[i])
        sum_ends.append(ends[i] if ends else starts[i])
        sum_values.append(values[i])

        for sums_index in range(last_valid_index, len(sum_starts) - 1):
            if (
                ends
                and sum_ends[sums_index]
                and (sum_ends[sums_index] > ends[i] + timedelta(minutes=duration))
            ):
                last_valid_index += 1
                continue

            elif sum_ends[sums_index] and sum_ends[sums_index] > starts[i] + timedelta(
                minutes=duration
            ):
                last_valid_index += 1
                continue

            sum_starts[sums_index] = starts[i]
            if not sum_ends[sums_index]:
                sum_ends[sums_index] = ends[i] if ends else starts[i]
            sum_values[sums_index] += values[i]

    assert (
        len(sum_starts) == len(sum_ends) == len(sum_values)
    ), "expected output shapes to match"

    sum_starts.reverse()
    sum_ends.reverse()
    sum_values.reverse()

    return (sum_starts, sum_ends, sum_values)


def lists_sorter(list_1, list_2, list_3=None, list_4=None, list_5=None):
    """Sort lists that are matched index-wise, using the first list as the
        property to sort by

    Example:
        l1: [50, 2, 3]               ->     [2, 3, 50]
        l2: [dog, cat, parrot]       ->     [cat, parrot, dog]
    """
    unsort_1 = numpy.array(list_1)
    unsort_2 = numpy.array(list_2)
    unsort_3 = numpy.array(list_3)
    unsort_4 = numpy.array(list_4)
    unsort_5 = numpy.array(list_5)

    sort_indexes = unsort_1.argsort()

    unsort_1.sort()
    list_1 = list(unsort_1)
    l2 = list(unsort_2[sort_indexes])
    if list_3:
        l3 = list(unsort_3[sort_indexes])
    else:
        l3 = []
    if list_4:
        l4 = list(unsort_4[sort_indexes])
    else:
        l4 = []
    if list_5:
        l5 = list(unsort_5[sort_indexes])
    else:
        l5 = []

    return (list_1, l2, l3, l4, l5)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul  3 15:36:31 2019

@author: annaquinlan

Github URL: https://github.com/tidepool-org/LoopKit/blob/
57a9f2ba65ae3765ef7baafe66b883e654e08391/LoopKit/CarbKit/CarbStatus.swift
"""
# pylint: disable=R0913, R0914
from datetime import timedelta

from pyloopkit.date import time_interval_since
from pyloopkit import carb_math


def dynamic_carbs_on_board_helper(
        carb_start,
        carb_value,
        absorption_dict,
        observed_timeline,
        at_date,
        default_absorption_time,
        delay,
        delta,
        carb_absorption_time=None
        ):
    """
    Find partial COB for a particular carb entry *dynamically*

    Arguments:
    carb_start -- time of carb entry (datetime objects)
    carb_value -- grams of carbs eaten

    absorption_dict -- list of absorption information
                       (computed via map_)
    observed_timeline -- list of carb absorption info at various times
                         (computed via map_)

    at_date -- date to calculate the glucose effect (datetime object)

    default_absorption_time -- absorption time to use for unspecified
                               carb entries

    delay -- the time to delay the carb effect
    carb_absorption_time -- time carbs will take to absorb (mins)

    Output:
    Carbohydrate value (g)
    """

    # We have to have absorption info for dynamic calculation
    if (at_date < carb_start - timedelta(minutes=delta)
            or not absorption_dict
       ):
        return carb_math.carbs_on_board_helper(
            carb_start,
            carb_value,
            at_date,
            default_absorption_time,
            delay,
            carb_absorption_time
            )

    # Less than minimum observed; calc based on min absorption rate
    if observed_timeline and None in observed_timeline[0]:
        time = time_interval_since(at_date, carb_start) / 60 - delay
        estimated_date_duration = (
            time_interval_since(
                absorption_dict[5],
                absorption_dict[4]
                ) / 60
            + absorption_dict[6]
        )
        return carb_math.linear_unabsorbed_carbs(
            absorption_dict[2],
            time,
            estimated_date_duration
            )

    if (not observed_timeline  # no absorption was observed (empty list)
            or not observed_timeline[len(observed_timeline)-1]
            or at_date > observed_timeline[len(observed_timeline) - 1][1]
       ):
        # Predict absorption for remaining carbs, post-observation
        total = absorption_dict[3]  # these are the still-unabsorbed carbs
        time = time_interval_since(at_date, absorption_dict[5]) / 60
        absorption_time = absorption_dict[6]

        return carb_math.linear_unabsorbed_carbs(
            total,
            time,
            absorption_time
        )

    # There was observed absorption
    total = carb_value
    def partial_absorption(dict_):
        if dict_[1] > at_date:
            return 0
        return dict_[2]

    for dict_ in observed_timeline:
        total -= partial_absorption(dict_)

    return max(
        total,
        0
        )


def dynamic_absorbed_carbs(
        carb_start,
        carb_value,
        absorption_dict,
        observed_timeline,
        at_date,
        carb_absorption_time,
        delay,
        delta,
        ):
    """
    Find partial absorbed carbs for a particular carb entry *dynamically*

    Arguments:
    carb_start -- time of carb entry (datetime objects)
    carb_value -- grams of carbs eaten

    absorption_dict -- list of absorption information
                       (computed via map_)
    observed_timeline -- list of carb absorption info at various times
                         (computed via map_)

    at_date -- date to calculate the glucose effect (datetime object)

    carb_absorption_time -- time carbs will take to absorb (mins)

    delay -- the time to delay the carb effect

    Output:
    Carbohydrate value (g)
    """

    # We have to have absorption info for dynamic calculation
    if (at_date < carb_start
            or not absorption_dict
       ):
        return carb_math.absorbed_carbs(
            carb_start,
            carb_value,
            carb_absorption_time,
            at_date,
            delay,
            )
    # Less than minimum observed; calc based on min absorption rate
    if observed_timeline and None in observed_timeline[0]:
        time = time_interval_since(at_date, carb_start) / 60 - delay
        estimated_date_duration = (
            time_interval_since(
                absorption_dict[5],
                absorption_dict[4]
                ) / 60
            + absorption_dict[6]
        )
        return carb_math.linearly_absorbed_carbs(
            absorption_dict[2],
            time,
            estimated_date_duration
            )

    if (not observed_timeline  # no absorption was observed (empty list)
            or not observed_timeline[len(observed_timeline)-1]
            or at_date > observed_timeline[len(observed_timeline) - 1][1]
       ):
        # Predict absorption for remaining carbs, post-observation
        total = absorption_dict[3]  # these are the still-unabsorbed carbs
        time = time_interval_since(at_date, absorption_dict[5]) / 60
        absorption_time = absorption_dict[6]

        return absorption_dict[1] + carb_math.linearly_absorbed_carbs(
            total,
            time,
            absorption_time
        )

    sum_ = 0

    # There was observed absorption
    def filter_dates(sub_timeline):
        return sub_timeline[0] + timedelta(minutes=delta) <= at_date

    before_timelines = list(filter(filter_dates, observed_timeline))

    if before_timelines:
        last = before_timelines.pop()
        observation_interval = (last[1] - last[0]).total_seconds()
        if observation_interval > 0:
            # find the minutes of overlap between calculation_interval
            # and observation_interval
            calculation_interval = (
                last[1] - min(
                    last[0],
                    at_date
                    )
                ).total_seconds()
            sum_ += (calculation_interval
                     / observation_interval
                     * last[2]
                     )

    for dict_ in before_timelines:
        sum_ += dict_[2]

    return min(
        sum_,
        absorption_dict[0]
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyloopkit', 'pyloopkit.depreciated']

package_data = \
{'': ['*'],
 'pyloopkit': ['docs/*', 'example_files/*', 'tests/fixtures/LoopKit/*']}

install_requires = \
['numpy>=1.23.2,<2.0.0']

setup_kwargs = {
    'name': 'opsb-pyloopkit',
    'version': '0.1.0',
    'description': '',
    'long_description': "# PyLoopKit\nA set of Python tools for building closed-loop insulin delivery apps (Python port of LoopKit)\n\n[Link to Tidepool Loop repository version used for algorithm](https://github.com/tidepool-org/Loop/tree/8c1dfdba38fbf6588b07cee995a8b28fcf80ef69)\n\n[Link of Tidepool LoopKit repository version used for algorithm](https://github.com/tidepool-org/LoopKit/tree/57a9f2ba65ae3765ef7baafe66b883e654e08391)\n\n# To use this project\n## Please review [the documentation](pyloopkit/docs/pyloopkit_documentation.md) for usage instructions, input data requirements, and other important details.\n\n### To recreate the Virtual Environment\n1. This environment was developed with Anaconda. You'll need to install [Miniconda](https://conda.io/miniconda.html) or [Anaconda](https://anaconda-installer.readthedocs.io/en/latest/) for your platform.\n2. In a terminal, navigate to the directory where the environment.yml \nis located (likely in PyLoopKit/pyloopkit folder).\n3. Run `conda env create`; this will download all of the package dependencies\nand install them in a virtual environment named py-loop. PLEASE NOTE: this\nmay take up to 30 minutes to complete.\n\n### To use the Virtual Environment\nIn Bash run `source activate py-loop`, or in the Anaconda Prompt\nrun `conda activate py-loop` to start the environment.\n\nRun `deactivate` to stop the environment.\n\n### To create the PyLoopKit package\nIf you want to install a version of the PyLoopKit package based on the PyLoopKit on your local machine, run `python3 setup.py install` within your PyLoopKit repo to call the setup script and install the package. You'll likely want to do this within the `py-loop` environment.\n\n### Running the unittests\nTo run PyLoopKit's unit tests, run `python3 -m unittest discover` within your PyLoopKit repo\n",
    'author': 'Oliver Searle-Barnes',
    'author_email': 'oliver@opsb.co.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/tidepool-org/PyLoopKit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 21 13:11:40 2019

@author: annaquinlan, plot style from Ed Nykaza
"""
from collections import OrderedDict
from datetime import timedelta, datetime
import matplotlib.pyplot as plt
from matplotlib import collections as mc

from pyloopkit.date import time_interval_since
from pyloopkit.insulin_math import schedule_offset


def plot_graph(
        dates, values,
        x_label=None, y_label=None, title=None,
        line_color=None, fill_color=None,
        file_name=None, grid=False,
        line_style="-", scatter=False):
    """ Plot an Loop-style effects graph, with the x-axis ticks
        being absolute hour of the day (ex: 13:00)

    dates -- datetimes of dates to plot at
    values -- integer values to plot
    Optional parameters:
        x_label -- the x-axis label
        y_label -- the y-axis label
        title -- the title of the graph
        line_color -- color of the line that is graphed
        fill_color -- the color of the fill under the graph (defaults to
                      no fill)
        file_name -- name to save the plot as (if no name is specified, the
                     graph is not saved)
        grid -- set to True to enable a grid on the graph
        line_style -- see pyplot documentation for the line style options
        scatter -- plot points as a scatter plot instead of a line
    """
    relative_dates = []

    # convert from exact dates to relative dates
    for i in range(0, len(dates)):
        date = dates[i]
        hours = date.hour + date.minute / 60 + date.second / 3600

        # adjust if the times cross midnight
        if (len(relative_dates) > 0
                and relative_dates[i-1] > hours):
            hours += 24
        relative_dates.append(hours)

    font = {
        'family': 'DejaVu Sans',
        'weight': 'bold',
        'size': 15
    }
    plt.rc('font', **font)

    figure_size_inches = (15, 7)
    fig, ax = plt.subplots(figsize=figure_size_inches)

    coord_color = "#c0c0c0"
    ax.spines['bottom'].set_color(coord_color)
    ax.spines['top'].set_color(coord_color)
    ax.spines['left'].set_color(coord_color)
    ax.spines['right'].set_color(coord_color)
    ax.xaxis.label.set_color(coord_color)
    ax.tick_params(axis='x', colors=coord_color)
    ax.yaxis.label.set_color(coord_color)
    ax.tick_params(axis='y', colors=coord_color)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)

    x_ticks_duplicates = [date.hour for date in dates]
    x_ticks = list(OrderedDict.fromkeys(x_ticks_duplicates))
    labels = ["%d:00" % x1 for x1 in x_ticks]

    # if we cross midnight, adjust so the graph displays correctly
    if not sorted(x_ticks) == x_ticks:
        for i in range(1, len(x_ticks)):
            if x_ticks[i-1] > x_ticks[i]:
                x_ticks[i] = x_ticks[i] + 24

    plt.xticks(x_ticks, labels)

    if x_label:
        ax.set_xlabel(x_label)
    if y_label:
        ax.set_ylabel(y_label)
    if title:
        plt.title(title, loc="left", fontweight='bold')

    if scatter:
        ax.scatter(
            relative_dates, values, color=line_color or "#f09a37"
        )
    else:
        ax.plot(
            relative_dates, values, color=line_color or "#f09a37", lw=4,
            ls=line_style
        )
    if fill_color:
        plt.fill_between(
            relative_dates, values, color=fill_color or "#f09a37", alpha=0.5
        )
    if grid:
        plt.grid(grid)
    if file_name:
        plt.savefig(file_name + ".png")
    plt.show()


def plot_relative_graph(
        dates, values,
        x_label=None, y_label=None, title=None,
        line_color=None, fill_color=None, file_name=None, grid=False,
        line_style="-", scatter=False):
    """ Plot an Loop-style effects graph, with the x-axis ticks
        being the relative time since the first value (ex: 4 hours)

    dates -- datetimes of dates to plot at
    values -- integer values to plot

    Optional parameters:
        x_label -- the x-axis label
        y_label -- the y-axis label
        title -- the title of the graph
        line_color -- color of the line that is graphed
        fill_color -- the color of the fill under the graph (defaults to
                      no fill)
        file_name -- name to save the plot as (if no name is specified, the
                     graph is not saved)
        grid -- set to True to enable a grid on the graph
        line_style -- see pyplot documentation for the line style options
        scatter -- plot points as a scatter plot instead of a line
    """
    relative_dates = []

    # convert from exact dates to relative dates
    for date in dates:
        relative_dates.append(
            time_interval_since(date, dates[0]) / 3600
        )

    font = {
        'family': 'DejaVu Sans',
        'weight': 'bold',
        'size': 15
    }
    plt.rc('font', **font)

    figure_size_inches = (15, 7)
    fig, ax = plt.subplots(figsize=figure_size_inches)

    coord_color = "#c0c0c0"
    ax.spines['bottom'].set_color(coord_color)
    ax.spines['top'].set_color(coord_color)
    ax.spines['left'].set_color(coord_color)
    ax.spines['right'].set_color(coord_color)
    ax.xaxis.label.set_color(coord_color)
    ax.tick_params(axis='x', colors=coord_color)
    ax.yaxis.label.set_color(coord_color)
    ax.tick_params(axis='y', colors=coord_color)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)

    x_ticks_duplicates = [date.hour - dates[0].hour for date in dates]
    x_ticks = list(OrderedDict.fromkeys(x_ticks_duplicates))
    labels = ["%d" % x1 for x1 in x_ticks]
    plt.xticks(x_ticks, labels)

    if x_label:
        ax.set_xlabel(x_label)
    if y_label:
        ax.set_ylabel(y_label)
    if title:
        plt.title(title, loc="left", fontweight='bold')

    if scatter:
        ax.scatter(
            relative_dates, values, color=line_color or "#f09a37"
        )
    else:
        ax.plot(
            relative_dates, values, color=line_color or "#f09a37", lw=4,
            ls=line_style
        )
    if fill_color:
        plt.fill_between(
            relative_dates, values, color=fill_color or "#f09a37", alpha=0.5
        )
    if grid:
        plt.grid(grid)
    if file_name:
        plt.savefig(file_name + ".png")
    plt.show()


def plot_multiple_relative_graphs(
        dates, values,
        x_label=None, y_label=None, title=None,
        line_color=None, fill_color=None, file_name=None, grid=False):
    """ Plot an Loop-style effects graph, with the x-axis ticks
        being the relative time since the first value (ex: 4 hours) AND there
        being multiple lines on the same graph

    dates -- lists of dates to plot at (datetime)
        ex: [
                [1:00, 2:00],
                [1:00, 2:00]
            ]
    values -- lists of integer values to plot
        ex: ex: [
                [2, 4],
                [0, -20]
            ]

    Optional parameters:
        x_label -- the x-axis label
        y_label -- the y-axis label
        title -- the title of the graph
        line_color -- color of the line that is graphed
        fill_color -- the color of the fill under the graph (defaults to
                      no fill)
        file_name -- name to save the plot as (if no name is specified, the
                     graph is not saved)
        grid -- set to True to enable a grid on the graph
        line_style -- see pyplot documentation for the line style options
        scatter -- plot points as a scatter plot instead of a line
    """
    assert len(dates) == len(values)

    font = {
        'family': 'DejaVu Sans',
        'weight': 'bold',
        'size': 15
    }
    plt.rc('font', **font)

    figure_size_inches = (15, 7)
    fig, ax = plt.subplots(figsize=figure_size_inches)

    coord_color = "#c0c0c0"
    ax.spines['bottom'].set_color(coord_color)
    ax.spines['top'].set_color(coord_color)
    ax.spines['left'].set_color(coord_color)
    ax.spines['right'].set_color(coord_color)
    ax.xaxis.label.set_color(coord_color)
    ax.tick_params(axis='x', colors=coord_color)
    ax.yaxis.label.set_color(coord_color)
    ax.tick_params(axis='y', colors=coord_color)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)

    relative_dates = []
    # convert from exact dates to relative dates
    for date_list in dates:
        relative_date_list = []
        for date in date_list:
            relative_date_list.append(
                time_interval_since(date, date_list[0]) / 3600
            )
        relative_dates.append(relative_date_list)

    x_ticks_duplicates = [date.hour - dates[0][0].hour for date in dates[0]]
    x_ticks = list(OrderedDict.fromkeys(x_ticks_duplicates))
    labels = ["%d" % x1 for x1 in x_ticks]
    plt.xticks(x_ticks, labels)

    if x_label:
        ax.set_xlabel(x_label)
    if y_label:
        ax.set_ylabel(y_label)
    if title:
        plt.title(title, loc="left", fontweight='bold')

    for (date_list, value_list) in zip(dates, values):
        ax.plot(
            date_list, value_list, color=line_color or "#f09a37", lw=4, ls="-"
        )
    if fill_color:
        plt.fill_between(
            relative_dates, values, color=fill_color or "#f09a37", alpha=0.5
        )
    if grid:
        plt.grid(grid)
    if file_name:
        plt.savefig(file_name + ".png")
    plt.show()


def plot_loop_inspired_glucose_graph(
        overall_dates, overall_values,
        momentum_dates=None, momentum_values=None,
        insulin_dates=None, insulin_values=None,
        carb_dates=None, carb_values=None,
        retrospective_dates=None, retrospective_values=None,
        previous_glucose_dates=None, previous_glucose_values=None,
        x_label=None, y_label=None, title=None,
        line_color=None, file_name=None, grid=False,
        line_style="-", target_min=None, target_max=None,
        correction_range_starts=None, correction_range_ends=None,
        correction_range_mins=None, correction_range_maxes=None):
    """ Create a Loop-inspired prediction line graph.

    overall_dates -- dates of overall prediction
    overall_values -- BG values of overall prediction

    momentum_dates, insulin_dates, carb_dates, retrospective_dates - dates
        of BG prediction that includes only the named effect
    momentum_values, insulin_values, carb_values, retrospective_values - values
        of BG prediction that includes only the named effect

    previous_glucose_dates -- dates of previous CGM points
    previous_glucose_values -- values of previous CGM points

    Optional parameters:
        x_label -- the x-axis label
        y_label -- the y-axis label
        title -- the title of the graph
        line_color -- color of the line that is graphed
        fill_color -- the color of the fill under the graph (defaults to
                      no fill)
        file_name -- name to save the plot as (if no name is specified, the
                     graph is not saved)
        grid -- set to True to enable a grid on the graph
        line_style -- see pyplot documentation for the line style options

        target_min -- the minimum correction range target for the whole
                      duration of the prediction
        target_max -- the maximum correction range target for the whole
                      duration of the prediction

        * passing these will override the target_min and target_max *
        correction_range_starts -- start times for target ranges (datetime)
        correction_range_ends -- stop times for given target ranges (datetime)
        correction_range_mins -- the lower bounds of target ranges (mg/dL)
        correction_range_maxes -- the upper bounds of target ranges (mg/dL)
    """

    def plot_line(
            absolute_dates, values,
            line_color="#5ac6fa",
            style="--",
            thickness=4,
            label=None):
        relative_dates = []
        for i in range(0, len(absolute_dates)):
            date = absolute_dates[i]
            hours = date.hour + date.minute / 60 + date.second / 3600

            # adjust if the times cross midnight
            if (len(relative_dates) > 0
                    and relative_dates[i-1] > hours):
                hours += 24
            relative_dates.append(hours)

        ax.plot(
            relative_dates, values, color=line_color,
            ls=style, lw=thickness, label=label
        )

    font = {
        'family': 'DejaVu Sans',
        'weight': 'bold',
        'size': 15
    }
    plt.rc('font', **font)

    figure_size_inches = (15, 7)
    fig, ax = plt.subplots(figsize=figure_size_inches)

    coord_color = "#c0c0c0"
    ax.spines['bottom'].set_color(coord_color)
    ax.spines['top'].set_color(coord_color)
    ax.spines['left'].set_color(coord_color)
    ax.spines['right'].set_color(coord_color)
    ax.xaxis.label.set_color(coord_color)
    ax.tick_params(axis='x', colors=coord_color)
    ax.yaxis.label.set_color(coord_color)
    ax.tick_params(axis='y', colors=coord_color)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)

    if previous_glucose_dates:
        x_ticks_duplicates = [
            date.hour for date in previous_glucose_dates
            ] + [date.hour for date in overall_dates]

    else:
        x_ticks_duplicates = [date.hour for date in overall_dates]

    x_ticks = list(OrderedDict.fromkeys(x_ticks_duplicates))
    labels = ["%d:00" % x1 for x1 in x_ticks]

    # if the times cross midnight, adjust
    if not sorted(x_ticks) == x_ticks:
        for i in range(1, len(x_ticks)):
            if x_ticks[i-1] > x_ticks[i]:
                x_ticks[i] = x_ticks[i] + 24

    plt.xticks(x_ticks, labels)

    # if there is a specified target range, it's assumed it will be for the
    # whole duration of the graph
    if target_min and target_max:
        # find the correct length to fill the correction range
        fill_length = [
            x_ticks[0],
            x_ticks[-1] if x_ticks[0] != x_ticks[-1] else x_ticks[0] + 1
        ]

        # plot correction range
        if abs(target_max - target_min) > 0:
            ax.fill_between(
                fill_length,
                [target_min, target_min],
                [target_max, target_max],
                facecolor='#B5E7FF',
                lw=0
            )
            plt.plot(
                [], [],
                color='#B5E7FF',
                lw=10,
                label="Target Range: %d-%d" % (target_min, target_max)
            )
        else:
            plt.axhline(
                y=target_min,
                color='#B5E7FF',
                lw=3,
                label="Target Range: %d-%d" % (target_min, target_max)
            )

    elif correction_range_starts:
        start = min(
            previous_glucose_dates[0] or overall_dates[0],
            overall_dates[0]
        )
        end = max(
            start + timedelta(hours=1),
            overall_dates[-1]
        )
        (range_starts,
         range_ends,
         range_mins,
         range_maxes
         ) = correction_ranges_between(
             correction_range_starts,
             correction_range_ends,
             correction_range_mins,
             correction_range_maxes,
             start,
             end
             )

        for i in range(0, len(range_starts)):
            # find the fill length, adjusting if the dates cross midnight
            fill_start = x_ticks[0] if i == 0 else (
                range_starts[i].hour if range_starts[i].day == start.day
                else range_starts[i].hour + 24
            )
            fill_end = x_ticks[-1] if i == len(range_starts) - 1 else (
                range_ends[i].hour if range_ends[i].day == start.day
                else range_ends[i].hour + 24
            )
            fill_length = (
                fill_start,
                fill_end
            )
            # plot the range
            if abs(range_mins[i] - range_maxes[i]) > 0:
                ax.fill_between(
                    fill_length,
                    [range_mins[i], range_mins[i]],
                    [range_maxes[i], range_maxes[i]],
                    facecolor='#B5E7FF',
                    lw=0
                )
                plt.plot(
                    [], [],
                    color='#B5E7FF',
                    lw=10,
                    label="Target Range"
                )
            else:
                line = [
                    [
                        (fill_length[0], range_mins[i]),
                        (fill_length[1], range_mins[i])
                    ]
                ]
                lc = mc.LineCollection(line, colors='#B5E7FF', linewidths=3)
                ax.add_collection(lc)

    # set labels and title
    if x_label:
        ax.set_xlabel(x_label)
    if y_label:
        ax.set_ylabel(y_label)
    if title:
        plt.title(title, loc="left", fontweight='bold')

    scatter_dates = []
    line_dates = []

    # convert from exact dates to relative dates, adjusting for date changes
    if previous_glucose_dates:
        for i in range(0, len(previous_glucose_dates)):
            date = previous_glucose_dates[i]
            hours = date.hour + date.minute / 60 + date.second / 3600

            # adjust if the times cross midnight
            if (len(scatter_dates) > 0
                    and scatter_dates[i-1] > hours):
                hours += 24
            scatter_dates.append(hours)

    for i in range(0, len(overall_dates)):
        date = overall_dates[i]
        hours = date.hour + date.minute / 60 + date.second / 3600

        # adjust if the times cross midnight
        if (len(line_dates) > 0
                and line_dates[i-1] > hours):
            hours += 24
        line_dates.append(hours)

    if previous_glucose_dates:
        ax.scatter(
            scatter_dates, previous_glucose_values,
            color=line_color or "#5ac6fa",
            lw=4,
            ls="-",
            s=10,
            label="CGM Data"
        )

    # Plot the overall prediction line
    plot_line(
        overall_dates, overall_values,
        label="Predicted Glucose (Overall)"
    )

    # Plot each individual effect, if it exists
    if momentum_dates:
        plot_line(
            momentum_dates, momentum_values,
            line_color="#eb5905", thickness=5,
            label="Predicted Glucose (Momentum Only)"
        )
    if insulin_dates:
        plot_line(
            insulin_dates, insulin_values,
            line_color="#f29741", thickness=3,
            label="Predicted Glucose (Insulin Only)"
        )
    if carb_dates:
        plot_line(
            carb_dates, carb_values,
            line_color="#5FCB49", thickness=3,
            label="Predicted Glucose (Carb Only)"
        )
    if retrospective_dates:
        plot_line(
            retrospective_dates, retrospective_values,
            line_color="#4253ed", thickness=3,
            label="Predicted Glucose (RC Only)"
        )

    # add a legend
    leg = plt.legend(numpoints=1)
    for text in leg.get_texts():
        text.set_color('#606060')
        text.set_weight('normal')

    # add a grid
    plt.grid(grid)
    # save the output graph
    if file_name:
        plt.savefig(file_name + ".png")

    # display the eventual BG
    ax.text(
        max(ax.get_xlim()),
        max(ax.get_ylim()) + 4,
        "Eventually %d mg/dL" % overall_values[-1],
        horizontalalignment="right",
        size=15,
        color="#a6a5a2"
    )

    plt.show()


def correction_ranges_between(
        correction_range_starts, correction_range_ends,
        correction_range_mins, correction_range_maxes,
        start_date, end_date,
        repeat_interval=24
    ):
    """ Returns a slice of scheduled basal rates that occur between two dates

    Arguments:
    basal_start_times -- list of times the basal rates start at
    basal_rates -- list of basal rates(U/hr)
    basal_minutes -- list of basal lengths (in mins)
    start_date -- start date of the range (datetime obj)
    end_date -- end date of the range (datetime obj)
    repeat_interval -- the duration over which the rates repeat themselves
                       (24 hours by default)

    Output:
    Tuple in format (basal_start_times, basal_rates, basal_minutes) within
    the range of dose_start_date and dose_end_date
    """
    timezone_info = start_date.tzinfo
    if start_date > end_date:
        return ([], [], [])

    reference_time_interval = timedelta(
        hours=correction_range_starts[0].hour,
        minutes=correction_range_starts[0].minute,
        seconds=correction_range_starts[0].second
    )
    max_time_interval = (
        reference_time_interval
        + timedelta(hours=repeat_interval)
    )

    start_offset = schedule_offset(start_date, correction_range_starts[0])

    end_offset = (
        start_offset
        + timedelta(seconds=time_interval_since(end_date, start_date))
    )

    # if a dose is crosses days, split it into separate doses
    if end_offset > max_time_interval:
        boundary_date = start_date + (max_time_interval - start_offset)
        (start_times_1,
         end_times_1,
         mins_1,
         maxes_1,
         ) = correction_ranges_between(
             correction_range_starts,
             correction_range_ends,
             correction_range_mins,
             correction_range_maxes,
             start_date,
             boundary_date,
             repeat_interval=repeat_interval
             )
        (start_times_2,
         end_times_2,
         mins_2,
         maxes_2,
         ) = correction_ranges_between(
             correction_range_starts,
             correction_range_ends,
             correction_range_mins,
             correction_range_maxes,
             boundary_date,
             end_date,
             repeat_interval=repeat_interval
             )

        return (start_times_1 + start_times_2,
                end_times_1 + end_times_2,
                mins_1 + mins_2,
                maxes_1 + maxes_2
                )

    start_index = 0
    end_index = len(correction_range_starts)

    for (i, start_time) in enumerate(correction_range_starts):
        start_time = timedelta(
            hours=start_time.hour,
            minutes=start_time.minute,
            seconds=start_time.second
        )
        if start_offset >= start_time:
            start_index = i
        if end_offset < start_time:
            end_index = i
            break

    reference_date = start_date - start_offset
    reference_date = datetime(
        year=reference_date.year,
        month=reference_date.month,
        day=reference_date.day,
        hour=reference_date.hour,
        minute=reference_date.minute,
        second=reference_date.second,
        tzinfo=timezone_info
        )

    if start_index > end_index:
        return ([], [], [])

    (output_start_times,
     output_end_times,
     output_mins,
     output_maxes
     ) = ([], [], [], [])

    for i in range(start_index, end_index):
        end_time = (timedelta(
            hours=correction_range_starts[i+1].hour,
            minutes=correction_range_starts[i+1].minute,
            seconds=correction_range_starts[i+1].second) if i+1 <
                    len(correction_range_starts) else max_time_interval)

        output_start_times.append(
            reference_date + timedelta(
                hours=correction_range_starts[i].hour,
                minutes=correction_range_starts[i].minute,
                seconds=correction_range_starts[i].second
            )
        )
        output_end_times.append(reference_date + end_time)
        output_mins.append(correction_range_mins[i])
        output_maxes.append(correction_range_maxes[i])

    assert len(output_start_times) == len(output_end_times) ==\
        len(output_mins) == len(output_maxes),\
        "expected output shape to match"

    return (output_start_times, output_end_times, output_mins, output_maxes)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 14 09:53:33 2019

@author: annaquinlan

Github URL: https://github.com/tidepool-org/LoopKit/blob/
57a9f2ba65ae3765ef7baafe66b883e654e08391/LoopKit/GlucoseKit/GlucoseMath.swift
"""
# pylint: disable=C0103, R0913, R0914
# disable pylint errors for too many arguments/variables
import math
from datetime import timedelta

from pyloopkit.date import time_interval_since
from pyloopkit.loop_math import simulation_date_range_for_samples


def linear_regression(x_list, y_list):
    """ Calculates slope and intercept using linear regression
    This implementation is not suited for large datasets

    Arguments:
    tuples_list -- An array of tuples containing x and y values

    Output:
    A tuple of slope and intercept values
    """
    assert len(x_list) == len(y_list), "expected input shapes to match"
    sum_x = 0.0
    sum_y = 0.0
    sum_xy = 0.0
    sum_x_squared = 0.0
    sum_y_squared = 0.0
    count = len(x_list)

    for i in range(0, len(x_list)):  # pylint: disable=C0200
        x = x_list[i]
        y = y_list[i]
        sum_x += x
        sum_y += y
        sum_xy += x * y
        sum_x_squared += x * x
        sum_y_squared += y * y

    try:
        slope = (((count * sum_xy) - (sum_x * sum_y)) /
                 ((count * sum_x_squared) - (sum_x * sum_x)))

        # I didn't include the intercept because it was unused
    except ZeroDivisionError:
        return float('NaN')

    return slope


def is_calibrated(display_list):
    """ Checks if no calibration entries are present
    Runtime: O(n)

    Arguments:
    obj_list -- list of Glucose-related objects with is_display_only
                property

    Output:
    Whether the collection contains calibration entries
    """
    return True in display_list


def is_continuous(date_list, delta=5):
    """ Checks whether the collection can be considered continuous

    Arguments:
    date_list -- list of datetime objects
    delta -- the (expected) time interval between CGM values

    Output:
    Whether the collection is continuous
    """
    try:
        return (
            abs(time_interval_since(date_list[0], date_list[-1]))
            < delta * (len(date_list)) * 60
        )

    except IndexError:
        print("Out of bounds error: list doesn't contain date values")
        return False


def has_single_provenance(provenance_list):
    """ Checks whether the collection is all from the same source
    Runtime: O(n)

    Arguments:
    provenance_list -- list of Glucose-related objects with
                provenance_identifier property

    Output:
    True if the samples are from same source
    """
    try:
        first_provenance = provenance_list[0]

    except IndexError:
        print("Out of bounds error: list doesn't contain objects")

    for provenance in provenance_list:
        if provenance != first_provenance:
            return False

    return True


def linear_momentum_effect(
        date_list, glucose_value_list, display_list, provenance_list,
        duration=30,
        delta=5
    ):
    """ Calculates the short-term predicted momentum effect using
        linear regression

    Arguments:
    date_list -- list of datetime objects
    glucose_value_list -- list of glucose values (unit: mg/dL)
    display_list -- list of display_only booleans
    provenance_list -- list of provenances (Strings)
    duration -- the duration of the effects
    delta -- the time differential for the returned values

    Output:
    tuple with format (date_of_glucose_effect, value_of_glucose_effect)
    """
    assert len(date_list) == len(glucose_value_list) == len(display_list)\
        == len(provenance_list), "expected input shape to match"

    if (len(date_list) <= 2 or not is_continuous(date_list)
            or is_calibrated(display_list)
            or not has_single_provenance(provenance_list)
       ):
        return ([], [])

    first_time = date_list[0]
    last_time = date_list[-1]
    (start_date, end_date) = simulation_date_range_for_samples(
        [last_time], [], duration, delta
    )

    def create_times(time):
        return abs(time_interval_since(time, first_time))

    slope = linear_regression(
        list(map(create_times, date_list)), glucose_value_list
    )

    if math.isnan(slope) or math.isinf(slope):
        return ([], [])

    date = start_date
    momentum_effect_dates = []
    momentum_effect_values = []

    while date <= end_date:
        value = (max(0, time_interval_since(date, last_time)) * slope)
        momentum_effect_dates.append(date)
        momentum_effect_values.append(value)
        date += timedelta(minutes=delta)

    assert len(momentum_effect_dates) == len(momentum_effect_values),\
        "expected output shape to match"
    return (momentum_effect_dates, momentum_effect_values)


def counteraction_effects(
        dates, glucose_values, displays, provenances,
        effect_dates, effect_values
    ):
    """ Calculates a timeline of effect velocity (glucose/time) observed
        in glucose readings that counteract the specified effects.

    Arguments:
    dates -- list of datetime objects of dates of glucose values
    glucose_values -- list of glucose values (unit: mg/dL)
    displays -- list of display_only booleans
    provenances -- list of provenances (Strings)

    effect_dates -- list of datetime objects associated with a glucose effect
    effect_values -- list of values associated with a glucose effect

    Output:
    An array of velocities describing the change in glucose samples
    compared to the specified effects
    """
    assert len(dates) == len(glucose_values) == len(displays)\
        == len(provenances), "expected input shape to match"
    assert len(effect_dates) == len(effect_values),\
        "expected input shape to match"

    if not dates or not effect_dates:
        return ([], [], [])

    effect_index = 0
    start_glucose = glucose_values[0]
    start_date = dates[0]
    start_prov = provenances[0]
    start_display = displays[0]

    start_dates = []
    end_dates = []
    velocities = []

    for i in range(1, len(dates)):
        # Find a valid change in glucose, requiring identical
        # provenance and no calibration
        glucose_change = glucose_values[i] - start_glucose
        time_interval = time_interval_since(dates[i], start_date)

        if time_interval <= 4 * 60:
            continue

        if (not start_prov == provenances[i]
                or start_display
                or displays[i]
           ):
            start_glucose = glucose_values[i]
            start_date = dates[i]
            start_prov = provenances[i]
            start_display = displays[i]
            continue

        start_effect_date = None
        start_effect_value = None
        end_effect_date = None
        end_effect_value = None

        for j in range(effect_index, len(effect_dates)):
            # if one of the start_effect properties doesn't exist and
            # the glucose effect at position "j" will happen after the
            # starting glucose date, then make start_effect equal to that
            # effect
            if (not start_effect_date
                    and effect_dates[j] >= start_date
               ):
                start_effect_date = effect_dates[j]
                start_effect_value = effect_values[j]

            elif (not end_effect_date
                  and effect_dates[j] >= dates[i]
                 ):
                end_effect_date = effect_dates[j]
                end_effect_value = effect_values[j]
                break

            effect_index += 1

        if end_effect_value is None:
            continue
        effect_change = end_effect_value - start_effect_value

        discrepancy = glucose_change - effect_change

        average_velocity = discrepancy / time_interval * 60

        start_dates.append(start_date)
        end_dates.append(dates[i])
        velocities.append(average_velocity)

        start_glucose = glucose_values[i]
        start_date = dates[i]
        start_prov = provenances[i]
        start_display = displays[i]

    assert len(start_dates) == len(end_dates) == len(velocities),\
        "expected output shape to match"
    return (start_dates, end_dates, velocities)
